# ==== HC_M3.py : cryptanalysis de l'Enigma via HC

import sys
import eniM3
import string	
from bib_util import *
import HC_SCORE
import getopt
from WALHS import *

ALPHA = string.ascii_uppercase

# ===================================
def dechiffre(msg, e ):
    crypto = ""
    for car in msg:
        crypto += e.encode( car )
    return crypto
# ===================================
def affi_score( valeur, lg):
    if valeur > 1:
        valeur = valeur / lg
    ch = ("%6.4f" % valeur)
    return ch
# ===================================
def extrait( debut, fin ):
    ALPHA = string.ascii_uppercase
    ch_extrait = ALPHA[ ALPHA.find(debut) : ALPHA.find(fin) + 1 ]
    return ch_extrait
# ===================================
def decode_grund( ch ):
    tb = ch.split(":")
    if len(tb) == 1:
        return (ch[0],ch[1],ch[2])
    gr_l = extrait( tb[0][0], tb[1][0])
    gr_m = extrait( tb[0][1], tb[1][1])
    gr_r = extrait( tb[0][2], tb[1][2])
    return (gr_l, gr_m, gr_r)
# ===================================
def decode_ring( ch ):
    tb = ch.split(":")
    if len(tb) == 1:
        return (ch[0], ch[1], ch[2])
    r_l = extrait( tb[0][0], tb[1][0])
    r_m = extrait( tb[0][1], tb[1][1])
    r_r = extrait( tb[0][2], tb[1][2])
    return (r_l, r_m, r_r)
# ===================================
def usage():
    print("usage:")
    print("  -h             Aide en ligne")
    print("  -c crypto      The cryptogram")
    print("  -v             Verbose (mode Debug)")
    print("  -W XYZ         The Walzenlage, 123 by default")
    print("  -G XYZ         The Grundstellung, AAA by default")
    print("  -R XYZ         The Ringstellungi, AAA by default")
    print("  -S AF:CD:PN    The Steckers, by default no steckers")
    print("  -M method      The method (BIG,TRI,...), default: ADE")
    print("  -m file_method The file used with method")
    print("  -U UKW         Reflector: 10(A),0 or 11(B),9 or 12(C), by default B")
# ------------------ DEBUT PROGRAMME ---------------------
# ----- on traite les options
try:
    opts, args = getopt.getopt(sys.argv[1:],
        "hvc:W:G:R:S:M:m:U:", ["help", "verbose", "crypto=", 
        "walzenlage=", "grund=", "ring=", "steckers=", "method=",
        "file_method=", "ukw=" ])
except getopt.GetoptError as err:
    usage()
    sys.exit(2)

DEBUG=False
msg = ""
WALZENLAGE = "123"
STECKS = ""
METHOD = "IC"
FILE_METHOD = "IC"
UKW = 0
GRUND = "AAA"
RING = "AAA"

for o, a in opts:
    if o == "-v":
        DEBUG = True
    elif o in ("-h", "--help"):
        usage()
        sys.exit()
    elif o in ("-c", "--crypto"):
        msg = litfic( a )
    elif o in ("-W", "--walhenzage"):
        WALZENLAGE = a
    elif o in ("-G", "--grund"):
        GRUND = a
    elif o in ("-R", "--ring"):
        RING = a
    elif o in ("-S", "--steckers"):
        STECKS = a
    elif o in ("-M", "--method"):
        METHOD = a 
    elif o in ("-m", "--file_method"):
        FILE_METHOD = a
    elif o in ("-U", "--ukw"):
        UKW = int(a)
    else:
        assert False, "unhandled option"
if msg == "":
    usage()
    sys.exit(1)
# ---------------------------
HC_SCORE.score_init[ METHOD ]( FILE_METHOD )
MAX_SCORE = 0
nb_dechif = 0
LEFT = int( WALZENLAGE[0] )
MIDDLE = int( WALZENLAGE[1] )
RIGHT = int( WALZENLAGE[2] )

# -- On calcule score sans Steckers
c = eniM3.Enigma(UKW,LEFT,MIDDLE,RIGHT, STECKS, RING, GRUND, DEBUG)
nb_dechif += 1
plain = dechiffre( msg, c )
score = HC_SCORE.score[ METHOD ](plain)
MAX_SCORE = score
SCORE_SEUIL = score
MAX_PLAIN = plain
MAX_STECK = STECKS
CURRENT_STECKS = STECKS
print("UKW:",UKW," Walzenlage: ", WALZENLAGE, end=" ")
print(" Ringstellung: ", RING, " Grundstellung:", GRUND, end=" ")
print(" Steckers: ", STECKS)
print(affi_score( MAX_SCORE, len(msg) ),\
   "-", MAX_PLAIN[0:50], STECKS, nb_dechif)
# -- On flags les steckers donnees en arguments
flag = [0]*26
les_stecks = STECKS.split(":")
if les_stecks[0] != "":
 for un_steck in les_stecks:
    idx1 = ALPHA.find(un_steck[0])
    idx2 = ALPHA.find(un_steck[1])
    flag[idx1] = 1
    flag[idx2] = 1
# -- On calcule avec 4 steckers
for i in range(10):
 for indx_a in range(26):
    if flag[ indx_a ]: continue
    for indx_b in range(26):
        # -- on permutte
        if flag[ indx_b ]: continue
        if indx_a == indx_b : continue
        lettre_a = string.ascii_uppercase[ indx_a ]
        lettre_b = string.ascii_uppercase[ indx_b ]
        A_STECK = lettre_a + lettre_b
        if STECKS == "":
            CURRENT_STECKS = A_STECK
        else:
            CURRENT_STECKS = STECKS + ":" + A_STECK
        #print A_STECK,
        #print CURRENT_STECKS

        # -- on chiffre
        c = eniM3.Enigma(UKW,LEFT,MIDDLE,RIGHT, CURRENT_STECKS, \
            RING, GRUND, DEBUG)
        nb_dechif += 1
        plain = dechiffre( msg, c )
        score = HC_SCORE.score[ METHOD ](plain)

        # -- on memorise si c'est le meilleur score
        if score > MAX_SCORE:
            MAX_SCORE = score
            MAX_PLAIN = plain
            MAX_A_STECK = A_STECK
            MAX_indxa = indx_a ; MAX_indxb = indx_b
 if MAX_SCORE > SCORE_SEUIL:
  print(i, affi_score( MAX_SCORE, len(msg) ),\
    "-", MAX_PLAIN[0:50], MAX_A_STECK, nb_dechif)
  flag[ MAX_indxa ] = 1 ; flag[ MAX_indxb ] = 1 
  SCORE_SEUIL = MAX_SCORE
  if STECKS == "":
    STECKS = MAX_A_STECK
  else:
    STECKS += ":" + MAX_A_STECK
 else:
  break   
print("nombre de dechiffrements: ", nb_dechif)
print("STECKERS: ", STECKS)

