# ==== HC_M3v2.py : cryptanalysis de l'Enigma via HC
# J'utilise l'algorithme de Frode Weirund

import sys
import eniM3
import string	
from bib_util import *
import HC_SCORE
import getopt
import random
from WALHS import *

ALPHA = string.ascii_uppercase

# ===================================
def gene_stecks(steck_alpha):
    steckers = ""
    for i in range(26):
        if steck_alpha[i] == ALPHA[i]: continue
        if steck_alpha[i] < ALPHA[i]: continue
        a_steck = ALPHA[i] + steck_alpha[i]
        if steckers == "":
            steckers = a_steck
        else:
            steckers += ":" + a_steck
    return steckers
# ===================================
def verif_stecks(steck_alpha):
    for i in range(len(steck_alpha)):
        j = ALPHA.find(steck_alpha[i])
        if steck_alpha[j] != ALPHA[i]:
            print("=======================> ERREUR")
            sys.exit(1)
            return False
    return True
# ===================================
def change_stecks(steck_alpha, i, j):
    lst = list(steck_alpha)
    tmp = lst[i]
    lst[i] = lst[j]
    lst[j] = tmp
    return "".join(lst)
# ===================================
def dechiffre(msg, e ):
    crypto = ""
    for car in msg:
        crypto += e.encode( car )
    return crypto
# ===================================
def affi_score( valeur, lg):
    if valeur > 1:
        valeur = valeur / lg
    ch = ("%6.4f" % valeur)
    return ch
# ===================================
def extrait( debut, fin ):
    ALPHA = string.ascii_uppercase
    ch_extrait = ALPHA[ ALPHA.find(debut) : ALPHA.find(fin) + 1 ]
    return ch_extrait
# ===================================
def decode_grund( ch ):
    tb = ch.split(":")
    if len(tb) == 1:
        return (ch[0],ch[1],ch[2])
    gr_l = extrait( tb[0][0], tb[1][0])
    gr_m = extrait( tb[0][1], tb[1][1])
    gr_r = extrait( tb[0][2], tb[1][2])
    return (gr_l, gr_m, gr_r)
# ===================================
def decode_ring( ch ):
    tb = ch.split(":")
    if len(tb) == 1:
        return (ch[0], ch[1], ch[2])
    r_l = extrait( tb[0][0], tb[1][0])
    r_m = extrait( tb[0][1], tb[1][1])
    r_r = extrait( tb[0][2], tb[1][2])
    return (r_l, r_m, r_r)
# ===================================
def usage():
    print("usage:")
    print("  -h             Aide en ligne")
    print("  -c crypto      The cryptogram")
    print("  -v             Verbose (mode Debug)")
    print("  -W XYZ         The Walzenlage, 123 by default")
    print("  -G XYZ         The Grundstellung, AAA by default")
    print("  -R XYZ         The Ringstellungi, AAA by default")
    print("  -M method[,...]The method(s) (BIG,TRI,...), default: IC")
    print("  ex:  -M BIG -M IC,BIG -M IC,IC :  we can have double pass")
    print("  -m file_method The file used with method")
    print("  -U UKW         Reflector: 10(A),0 or 11(B),9 or 12(C), by default B")
# ------------------ DEBUT PROGRAMME ---------------------
# ----- on traite les options
try:
    opts, args = getopt.getopt(sys.argv[1:],
        "hvc:W:G:R:M:m:U:", ["help", "verbose", "crypto=", 
        "walzenlage=", "grund=", "ring=", "method=",
        "file_method=", "ukw=" ])
except getopt.GetoptError as err:
    usage()
    sys.exit(2)

DEBUG=False
msg = ""
WALZENLAGE = "123"
STECKS = ""
METHODS = ["IC"]
FILE_METHOD = "IC"
UKW = 0
GRUND = "AAA"
RING = "AAA"

for o, a in opts:
    if o == "-v":
        DEBUG = True
    elif o in ("-h", "--help"):
        usage()
        sys.exit()
    elif o in ("-c", "--crypto"):
        msg = litfic( a )
    elif o in ("-W", "--walhenzage"):
        WALZENLAGE = a
    elif o in ("-G", "--grund"):
        GRUND = a
    elif o in ("-R", "--ring"):
        RING = a
    elif o in ("-M", "--method"):
        METHODS = a.split(",")
    elif o in ("-m", "--file_method"):
        FILE_METHOD = a
    elif o in ("-U", "--ukw"):
        UKW = int(a)
    else:
        assert False, "unhandled option"
if msg == "":
    usage()
    sys.exit(1)
# ---------------------------
METHOD = METHODS[0]
HC_SCORE.score_init[ METHOD ]( FILE_METHOD )
MAX_SCORE = 0
nb_dechif = 0
LEFT = int( WALZENLAGE[0] )
MIDDLE = int( WALZENLAGE[1] )
RIGHT = int( WALZENLAGE[2] )

# -- On calcule score sans Steckers
c = eniM3.Enigma(UKW,LEFT,MIDDLE,RIGHT, STECKS, RING, GRUND, 0)
nb_dechif += 1
plain = dechiffre( msg, c )
score = HC_SCORE.score[ METHOD ](plain)
MAX_SCORE = score
SCORE_SEUIL = score
MAX_PLAIN = plain
MAX_STECK = STECKS
print("UKW:",UKW," Walzenlage: ", WALZENLAGE, end=" ")
print(" Ringstellung: ", RING, " Grundstellung:", GRUND, end=" ")
print(" Steckers: ", STECKS)
print(affi_score( MAX_SCORE, len(msg) ),\
   "-", MAX_PLAIN[0:50], STECKS, nb_dechif)

# -- On teste tous les permutations possibles
STECKERS = ALPHA
for k in range(len(METHODS)):
 METHOD = METHODS[k]
 HC_SCORE.score_init[ METHOD ]( FILE_METHOD )
 for i in range(26-1):
  for j in range(i,26):
    idxi1 = ALPHA.find(ALPHA[i])
    idxi2 = ALPHA.find(STECKERS[i])
    idxj1 = ALPHA.find(ALPHA[j])
    idxj2 = ALPHA.find(STECKERS[j])

    LES_STECKS = []
    if   ( (idxi1 == idxi2 and idxj1 == idxj2) or \
           (idxi1 == idxj2 and idxj1 == idxi2)): 
        STECKERS_N1 = change_stecks(STECKERS, i, j)
        LES_STECKS.append(STECKERS_N1)
    elif ( (idxi1 == idxi2 and idxj1 != idxj2)):
        z = ALPHA.find(STECKERS[j])
        STECKERS_N1 = change_stecks(STECKERS, j, z)
        STECKERS_N1 = change_stecks(STECKERS_N1, i, j)
        STECKERS_N2 = change_stecks(STECKERS_N1, i, j)
        STECKERS_N2 = change_stecks(STECKERS_N2, i, z)
        LES_STECKS.append(STECKERS_N1)
        LES_STECKS.append(STECKERS_N2)
    elif ( (idxj1 == idxj2 and idxi1 != idxi2)):
        z = ALPHA.find(STECKERS[i])
        STECKERS_N1 = change_stecks(STECKERS, i, z)
        STECKERS_N1 = change_stecks(STECKERS_N1, j, i)
        STECKERS_N2 = change_stecks(STECKERS_N1, j, i)
        STECKERS_N2 = change_stecks(STECKERS_N2, j, z)
        LES_STECKS.append(STECKERS_N1)
        LES_STECKS.append(STECKERS_N2)
    elif ( (idxi1 != idxi2 and idxj1 != idxj2)):
        x = ALPHA.find(STECKERS[i])
        y = ALPHA.find(STECKERS[j])
        STECKERS_N1 = change_stecks(STECKERS, i, x)
        STECKERS_N1 = change_stecks(STECKERS_N1, j, y)
        STECKERS_N1 = change_stecks(STECKERS_N1, i, j)
        STECKERS_N2 = change_stecks(STECKERS_N1, x, y)
        STECKERS_N3 = change_stecks(STECKERS_N2, x, y)
        STECKERS_N3 = change_stecks(STECKERS_N3, i, j)
        STECKERS_N3 = change_stecks(STECKERS_N3, i, y)
        STECKERS_N4 = change_stecks(STECKERS_N3, x, j)
        LES_STECKS.append(STECKERS_N1)
        LES_STECKS.append(STECKERS_N2)
        LES_STECKS.append(STECKERS_N3)
        LES_STECKS.append(STECKERS_N4)

    for des_stecks in LES_STECKS:
        # -- on convertie STECKERS au format de mon Enigma
        CURRENT_STECKS = gene_stecks( des_stecks )

        # -- on chiffre
        c = eniM3.Enigma(UKW,LEFT,MIDDLE,RIGHT, CURRENT_STECKS, \
            RING, GRUND, 0)
        nb_dechif += 1
        plain = dechiffre( msg, c )
        score = HC_SCORE.score[ METHOD ](plain)

        # -- on memorise si c'est le meilleur score
        if score > MAX_SCORE:
            MAX_SCORE = score
            MAX_PLAIN = plain
            MAX_STECKS = des_stecks
            STECKERS = des_stecks
            print(i, affi_score( MAX_SCORE, len(msg) ),\
                "-", MAX_PLAIN[0:50], gene_stecks(MAX_STECKS), nb_dechif)

print("nombre de dechiffrements: ", nb_dechif)
print("MAX MAX SCORE: ", MAX_SCORE, STECKERS, MAX_PLAIN)

